package Active;

/**
	ActivePrime -- implements a prime as an active object
	
	Can we be sure that primes will be linked in the correct order?
	[Primes are passed along in a strictly ordered stream, and can
	only be approved by the first prime whose square is bigger
	than the value tested => must be linked in order ...]
*/

class ActivePrime extends Thread implements IntSource{

	// class variables
	private static IntSource lastPrime;		// where to link the next prime

	// instance variables
	private int value;				// value of this prime
	private int square;				// square of this prime
	private IntSource intSrc;		// source of ints to test
	private boolean debugOn;		// debug printing

	private OneSlotBuffer slot;		// to pass test value to next ActivePrime

	// constructor
	public ActivePrime(int value, IntSource intSrc)
		throws ActivePrimeFailure
	{
		this(value, intSrc, false); // debugging is off
		// this(value, intSrc, true); // debugging is on
	}

	public ActivePrime(int value, IntSource intSrc, boolean debug)
		throws ActivePrimeFailure
	{
		this.value = value;
		this.square = value*value;
		this.intSrc = intSrc;
		this.debugOn = debug;
		slot = new OneSlotBuffer();
		lastPrime = this;			// NB: set class variable
		this.checkInvariant();		// Perform sanity check!
		System.out.print(value + "\n");
		System.out.flush();
		this.start();
	}

	public int value() {
		return this.value;
	}

	// primes must be linked in increasing order!
	// This method checks the invariant and (presumably never)
	// throws an exception if the invariant is violated.
	private void checkInvariant()
		throws ActivePrimeFailure
	{
		if (this.value > 2) {
			// If my value is not 2; then intSrc is also an ActivePrime
			int previous = ((ActivePrime) this.intSrc).value();
			if (previous > this.value) {
				throw new ActivePrimeFailure("Fatal error: prime "
					+ this.value + " linked after " + previous);
			}
		}
	}

	public void run() {
		debug("starting thread");
		int testValue = intSrc.getInt();	// may block
		while (testValue != 0) {
			debug("testing " + testValue);
			if (testValue < this.square) {
				debug("found new prime " + testValue);
				try {
					new ActivePrime(testValue, lastPrime, debugOn);
				} catch (Exception e) {
					System.err.println(e.getMessage());
					testValue = 0; // stop the thread
				}
			} else if ((testValue % this.value) > 0) {
				debug("passing on " + testValue);
				// passed test, so prepare to give to next prime
				this.putInt(testValue);
			} else {
				debug("rejecting " + testValue);
			}
			testValue = intSrc.getInt();	// may block
		}
		putInt(0); // stop condition
		debug("stopping thread");
	}

	private void debug(String msg) {
		if (debugOn) {
			System.out.println("DEBUG(" + this.value + "): " + msg);
		}
	}

	private void putInt(int val) {
		slot.put(new Integer(val));
		// System.out.println("TestSlot got " + val);
	}

	// ActivePrime implements the IntSource interface
	public int getInt() {
		int rval;
		rval = ((Integer) slot.get()).intValue();
		// System.out.println("TestSlot returns " + rval);
		return rval;
	}
}
